/**
 * \brief	Belichtungstimer
 *
 * \author	Fabian Greif <fabian.greif@rwth-aachen.de>
 * \date	4.8.2007
 */
// ----------------------------------------------------------------------------

#include <avr/io.h>
#include <avr/interrupt.h>

#include <util/delay.h>
#include <stdlib.h>

// ----------------------------------------------------------------------------

typedef	struct {
	uint8_t minutes;
	uint8_t seconds;
} time_t;

typedef int16_t timestamp_t;

#define	A	0x02
#define	B	0x08
#define	C	0x40
#define	D	0x80
#define	E	0x20
#define	F	0x01
#define	G	0x04
#define	DP	0x10

uint8_t _bit_mask[16] = {
	A|B|C|D|E|F,	// 0
	B|C,			// 1
	A|B|G|E|D,		// 2
	A|B|C|D|G,		// 3
	B|C|F|G,		// 4
	A|C|D|F|G,		// 5
	A|C|D|E|F|G,	// 6
	A|B|C,			// 7
	A|B|C|D|E|F|G,	// 8
	A|B|C|D|F|G,	// 9
	A|B|C|E|F|G,	// a
	C|D|E|F|G,		// b
	A|D|E|F,		// c
	B|C|D|E|G,		// d
	A|D|E|F|G,		// e
	A|E|F|G,		// f
};

uint8_t segment[4];

#define	OFF		0xff

#define	SW1		1
#define	SW2		2
#define	SW3		4

#define REPEAT_MASK		(SW1 | SW2)		// repeat: key 1, 2
#define REPEAT_START	200				// after x ms
#define REPEAT_NEXT		50				// every x ms

#define REPEAT_MASK2	SW3				// repeat: key 3
#define REPEAT_START2	300				// after x ms
#define REPEAT_NEXT2	100				// every x ms

char key_state;				// debounced and inverted key state:
							// bit = 1: key pressed
char key_press;				// key press detect
char key_rpt;				// key long press and repeat

volatile uint8_t flag = 0;

// ----------------------------------------------------------------------------
// Timer 0 Interrupt

ISR(TIMER0_COMPA_vect)
{
	static uint8_t counter = 0;
	static char ct0, ct1, rpt = 0;
	static uint16_t rpt2 = 0;
	char i;
	
	TCNT0 = 0;
	
	// shutdown led-driver
	PORTD &= ~((1<<PD3)|(1<<PD4)|(1<<PD5)|(1<<PD6));
	
	// wait a little bit until they are realy down
	_delay_us(2);
	
	// set new value
	PORTB = segment[counter];
	
	// select next segment
	counter++;
	if (counter == 1)
		PORTD |= (1<<PD5);
	else if (counter == 2)
		PORTD |= (1<<PD6);
	else if (counter == 3)
		PORTD |= (1<<PD4);
	else if (counter > 3)
	{
		// read switch values
		uint8_t _button = 0x7;
		uint8_t temp = PORTB;
		
		PORTA |= (1<<PA1);
		
		PORTB = (1<<PB2)|(1<<PB4);
		_delay_us(1);
		if (!(PINA & (1<<PA1)))
			_button &= ~1;
		
		PORTB = (1<<PB2)|(1<<PB3);
		_delay_us(1);
		if (!(PINA & (1<<PA1)))
			_button &= ~2;
		
		PORTB = (1<<PB3)|(1<<PB4);
		_delay_us(1);
		if (!(PINA & (1<<PA1)))
			_button &= ~4;
		
		PORTA &= ~(1<<PA1);
		
		// debounce keys
		i = key_state ^ ~_button;
		ct0 = ~(ct0 & i);
		ct1 = ct0 ^ (ct1 & i);
		i &= ct0 & ct1;
		key_state ^= i;
		key_press |= key_state & i;		// 0->1: key press detect
		
		if ((key_state & REPEAT_MASK) == 0)	// check repeat function
			rpt = REPEAT_START;		// start delay
		
		if (--rpt == 0) {
			rpt = REPEAT_NEXT;			// repeat delay
			key_rpt |= key_state & REPEAT_MASK;
		}
		
		if ((key_state & REPEAT_MASK2) == 0)	// check repeat function
			rpt2 = REPEAT_START2;		// start delay
		
		if (--rpt2 == 0) {
			rpt2 = REPEAT_NEXT2;			// repeat delay
			key_rpt |= key_state & REPEAT_MASK2;
		}
		
		PORTB  =  temp;
		PORTD |=  (1<<PD3);
		counter = 0;
	}
}

// ----------------------------------------------------------------------------
// Timer 1 Interrupt

ISR(TIMER1_COMPA_vect)
{
	TCNT1 = 0;
	flag = 1;
}

// ----------------------------------------------------------------------------
// check if a key has been pressed. Each pressed key is reported
// only once

static uint8_t get_key_press( uint8_t key_mask )
{
	cli();                                          // read and clear atomic !
	key_mask &= key_press;                          // read key(s)
	key_press ^= key_mask;                          // clear key(s)
	sei();
	return key_mask;
}

// ----------------------------------------------------------------------------
// check if a key has been pressed long enough such that the
// key repeat functionality kicks in. After a small setup delay
// the key is reported beeing pressed in subsequent calls
// to this function. This simulates the user repeadiately
// pressing and releasing the key.

static uint8_t get_key_rpt( uint8_t key_mask )
{
	cli();                                          // read and clear atomic !
	key_mask &= key_rpt;                            // read key(s)
	key_rpt ^= key_mask;                            // clear key(s)
	sei();
	return key_mask;
}

// ----------------------------------------------------------------------------

static uint8_t get_key_short( uint8_t key_mask )
{
	cli();                                          // read key state and key press atomic !
	return get_key_press( ~key_state & key_mask );
}

// ----------------------------------------------------------------------------

static uint8_t get_key_long( uint8_t key_mask )
{
	return get_key_press( get_key_rpt( key_mask ));
}

// ----------------------------------------------------------------------------

static uint8_t get_key_pressed(uint8_t key_mask)
{
	cli();
	key_mask &= key_state;
	sei();
	return key_mask;
}

// ----------------------------------------------------------------------------
// generate bit mask

static uint8_t generate_mask(uint8_t character)
{
	uint8_t mask;
	
	if (character <= 9)
		mask = _bit_mask[character];
	else if (character >= 'a' && character <= 'f')
		mask = _bit_mask[character -'a' + 10];
	else if (character == 's')
		mask = _bit_mask[5];
	else if (character == 't')
		mask = D|E|F|G;
	else if (character == 'o')
		mask = C|D|E|G;
	else if (character == 'p')
		mask = A|B|E|F|G;
	else
		mask = 0;
	
	return mask;
}

// ----------------------------------------------------------------------------
// functions to converts a timestamp to a time value and vice versa

static time_t timestamp_to_time(timestamp_t time)
{
	div_t d = div(time, 60);
	
	time_t time_;
	
	time_.minutes = d.quot;
	time_.seconds = d.rem;
	
	return time_;
}

static timestamp_t time_to_timestamp(time_t time) {
	return time.minutes * 60 + time.seconds;
}

// ----------------------------------------------------------------------------
// displays a time value

void display_time(time_t time)
{
	// display minutes
	div_t d = div(time.minutes, 10);
	
	if (d.quot == 0)
		segment[0] = 0;
	else
		segment[0] = generate_mask(d.quot);
	
	segment[1] = generate_mask(d.rem) | DP;
	
	// display seconds
	d = div(time.seconds, 10);
	
	segment[2] = generate_mask(d.quot);
	segment[3] = generate_mask(d.rem);
}

// ----------------------------------------------------------------------------
// main loop

int main(void)
{
	DDRA = (1 << PA0);
	PORTA = 0;
	
	DDRB = 0xff;
	PORTB = 0x00;
	
	DDRD = 0x7e;
	PORTD = 0x00;
	
	// initilise timer 0, ctc, prescaler = 64
	TCCR0A = WGM01;
	TCCR0B = (1 << CS01) | (1 << CS00);
	OCR0A = 64;
	
	// initilise timer 0, ctc, prescaler = 256
	TCCR1B = (1 << WGM12) | (1 << CS12);
	OCR1A = 31250;
	
	TIMSK = (1 << OCIE1A) | (1 << OCIE0A);
	
	time_t time;
	
	time.minutes = 0;
	time.seconds = 15;
	
	display_time(time);
	
	sei();		// enable interrupts
	
	timestamp_t timestamp = time_to_timestamp(time);
	timestamp_t time_left = 0;
	
	enum { SET_TIME, RUN, PAUSE, STOP } mode = SET_TIME;
	
	while (1)
	{
		
		switch (mode)
		{
			case SET_TIME:
				if (get_key_press(SW1) || get_key_rpt(SW1)) {
					if (get_key_pressed(SW3))
						timestamp += 60;
					else
						timestamp += 1;
					
					// check if timestamp exceed the maximum value
					if (timestamp > (99 * 60 + 59))
						timestamp = 99 * 60 + 59;
				}
				
				if (get_key_press(SW2) || get_key_rpt(SW2)) {
					if (get_key_pressed(SW3))
						timestamp -= 60;
					else
						timestamp -= 1;
					
					// check if timestamp is below the minimum
					if (timestamp < 0)
						timestamp = 0;
				}
				display_time(timestamp_to_time(timestamp));
				
				get_key_long(SW3);
				if (get_key_short(SW3)) {
					time_left = timestamp;
					mode = RUN;
					
					flag = 1;
					TCNT1 = 0;
				}
				break;
			
			case RUN:
				if (flag) {
					flag = 0;
					
					--time_left;
					if (time_left < 0) {
						// time is up => display 'Stop' 
						time_left = 0;
						mode = STOP;
					}
				}
				
				// display time which is left
				display_time(timestamp_to_time(time_left));
				
				if (get_key_press(SW2)) {
					mode = SET_TIME;	// cancel
				}
				
				if (get_key_press(SW3)) {
					mode = PAUSE;		// pause timer
				}
				break;
			
			case PAUSE:
				if (get_key_press(SW2)) {
					mode = SET_TIME;	// cancel
				}
				
				if (get_key_press(SW3)) {
					mode = RUN;			// resume
					flag = 0;
					TCNT1 = 0;
				}
				break;
			
			case STOP:
				if (time_left < 5) {
					segment[0] = generate_mask('s');
					segment[1] = generate_mask('t');
					segment[2] = generate_mask('o');
					segment[3] = generate_mask('p');
					
					if (flag) {
						flag = 0;
						
						time_left++;
						if (time_left >= 5) {
							segment[0] = 0;
							segment[1] = 0;
							segment[2] = 0;
							segment[3] = 0;
						}
					}
				}
				
				if (get_key_press(SW1) || get_key_press(SW2) || get_key_press(SW3))
					mode = SET_TIME;
				break;
				
			default:
				break;
		}
		
		if (mode == RUN) {
			PORTD |= (1 << PD1);
		}
		else {
			PORTD &= ~(1 << PD1);
		}
	}
	
	return 0;
}
